// Header File
// Created 11/21/2010; 6:58:28 AM

//******************************************************************************
//
// Project:         Jumpman 89
//
// file name:       jm_bullets.h
//
// author:          David Randall "Ranman" 
//                  derandall99@gmail.com
//
// notes::          Jumpman was originally created by Randy Glover for the
//                  Atari 400, Commodore 64, IBM PCjr, and Apple II in 1982
//
//******************************************************************************


#ifndef _jm_bullets_h
#define _jm_bullets_h

#include "BasicTypes.h"

#include "jm_enemies.h"

#define BULLET_STOP_SPEED 0
#define BULLET_NORM_SPEED 1
#define BULLET_FIRE_SPEED 3
#define BULLET_COUNT_MAX  8

typedef enum
{
	SCRIPT_ACTION_None           = 0,
	SCRIPT_ACTION_Visible        = 1,
	SCRIPT_ACTION_Invisible      = 2,
	SCRIPT_ACTION_Active         = 3,
	SCRIPT_ACTION_Inactive       = 4,
	SCRIPT_ACTION_Drop           = 5,
	SCRIPT_ACTION_Total
} SCRIPT_ACTION_Type;


typedef struct
{
	int2  x;                // current lcd x coord
	int2  y;                // starting lcd y coord
	int2  delay;            // frame delay count
	bool  active;           // bullet is active
	uint1 game_piece;       // GAME_PIECE_Type
	uint1 speed;            // speed of bullet;
	uint1 direction;        // DIR_Type
	uint1 anim_step;        // animation step for zig-zag and hailstones
	uint1 action;           // BULLET_ACTION_Type;
} BULLET_OBJECT_Type;


extern BULLET_OBJECT_Type bullet_list[BULLET_COUNT_MAX];
extern int2 bullet_count;


void DrawBullets(void);
void UpdateBullets(void);
void CheckBullets(void);
void InitBullets(void);
void ResetBullet(BULLET_OBJECT_Type* bullet);
void FireBullet(ENEMY_OBJECT_Type* enemy, GAME_PIECE_Type bullet_type, DIR_Type direction);


#endif
