// C Source File
// Created 11/21/2010; 6:36:54 AM

//******************************************************************************
//
// Project:         Jumpman 89
//
// file name:       jm_enemies.c
//
// author:          David Randall "Ranman" 
//                  derandall99@gmail.com
//
// notes::          Jumpman was originally created by Randy Glover for the
//                  Atari 400, Commodore 64, IBM PCjr, and Apple II in 1982
//
//******************************************************************************


#include <tigcclib.h>
#include "extgraph.h"

#include "BasicTypes.h"
#include "BasicUtils.h"
#include "GrayUtils.h"
#include "FontUtils.h"

#include "jm_enemies.h"
#include "jm_graphics.h"
#include "jm_jumpman.h"
#include "jm_bullets.h"
#include "jm_collectables.h"
#include "jm_levels.h"

extern bool collectable_trigger;

uint1 GetAIType3Move(ENEMY_OBJECT_Type* enemy);



ENEMY_OBJECT_Type enemy_list[ENEMY_COUNT_MAX];

const ENEMY_OBJECT_Type enemy_definition_list[] =
{
	{
		GAME_PIECE_Falling_Bomb,
		
		0,0,0,0,0,0,

		{{53,56,59,62},{0,0,0,0},{49,50,51,52},{0,0,0,0},{0,0,0,0}},
		{3,0,1,0,0},
		{4,0,4,0,0},
		{{0,0}}, // waypoint list populated during init by level data

		0,0,0,0,0,0,
		
		ENEMY_MOVE_SPEED,
		DIR_Down,
		0,
		AI_TYPE_Falling_Explode,
		0,
		0,
		0,

		false,
		false,
		true,
		true,
	},
	{
		GAME_PIECE_Ufo,
		
		0,0,0,0,0,0,

		{{45,45,47,47},{41,41,43,43},{41,41,43,43},{41,41,43,43},{41,41,43,43}},
		{2,2,2,2,2},
		{4,4,4,4,4},
		{{0,0}}, // waypoint list populated during init by level data
		
		0,0,0,0,0,0,

		ENEMY_MOVE_SPEED,
		DIR_Down,
		0,
		AI_TYPE_None,
		0,
		0,
		0,

		false,
		false,
		true,
		true,
	},
	{
		GAME_PIECE_Bat,
		
		0,0,0,0,0,0,

		{{26,26,26,26},{27,28,29,28},{27,28,29,28},{27,28,29,28},{27,28,29,28}},
		{1,1,1,1,1},
		{4,4,4,4,4},
		{{0,0}}, // waypoint list populated during init by level data
		
		0,0,0,0,0,0,

		ENEMY_MOVE_SPEED - 1,
		DIR_Down,
		0,
		AI_TYPE_None,
		0,
		0,
		0,

		true,
		false,
		false,
		true,
	},
	{
		GAME_PIECE_Fire,
		
		0,0,0,0,0,0,

		{{65,66,67,68},{65,66,67,68},{65,66,67,68},{65,66,67,68},{65,66,67,68}}, //  dead, up, down, left, right
		{1,1,1,1,1}, //  dead, up, down, left, right
		{4,4,4,4,4}, //  dead, up, down, left, right
		{{0,0}}, // waypoint list populated during init by level data
		
		0,0,0,0,0,0,

		ENEMY_STOP_SPEED,
		DIR_Down,
		0,
		AI_TYPE_None,
		0,
		0,
		0,

		true,
		false,
		false,
		true,
	},
	{
		GAME_PIECE_Bird,
		
		0,0,0,100,0,0,

		{{80,81,82,81},{80,81,82,81},{80,81,82,81},{80,81,82,81},{80,81,82,81}}, //  dead, up, down, left, right
		{1,1,1,1,1}, //  dead, up, down, left, right
		{4,4,4,4,4}, //  dead, up, down, left, right
		{{0,0}}, // waypoint list populated during init by level data
		
		0,0,0,0,0,0,

		ENEMY_MOVE_SPEED,
		DIR_Right,
		0,
		AI_TYPE_None,
		0,
		0,
		0,

		false,
		false,
		true,
		true,
	},
	{
		GAME_PIECE_Follower,
		
		0,0,0,0,0,0,

		{{0,0,0,0},{0,0,0,0},{0,0,0,0},{0,0,0,0},{0,0,0,0}}, // this is a special enemy, uses Jumpan sprite
		{1,1,1,1,1}, //  dead, up, down, left, right
		{1,1,1,1,1}, // this is a special enemy, uses Jumpan sprite
		{{0,0}}, // waypoint list populated during init by level data
		
		0,0,0,0,0,0,

		ENEMY_MOVE_SPEED,
		DIR_Down,
		0,
		AI_TYPE_None,
		0,
		0,
		0,

		true,
		false,
		true,
		true,
	},
	{
		GAME_PIECE_Robot_1,
		
		0,0,0,0,0,0,

		{{24,24,25,25},{24,24,25,25},{24,24,25,25},{22,22,23,23},{20,20,21,21}}, //  dead, up, down, left, right
		{1,1,1,1,1}, //  dead, up, down, left, right
		{4,4,4,4,4}, //  dead, up, down, left, right
		{{0,0}}, // waypoint list populated during init by level data
		
		0,0,0,0,0,0,

		ENEMY_MOVE_SPEED,
		DIR_Down,
		0,
		AI_TYPE_Walking_Waypoints,
		0,
		0,
		0,

		true,
		false,
		false,
		true,
	},
	{
		GAME_PIECE_Robot_2,
		
		0,0,0,0,0,0,

		{{96,96,97,97},{96,96,97,97},{96,96,97,97},{94,94,95,95},{92,92,93,93}}, //  dead, up, down, left, right
		{1,1,1,1,1}, //  dead, up, down, left, right
		{4,4,4,4,4}, //  dead, up, down, left, right
		{{0,0}}, // waypoint list populated during init by level data
		
		0,0,0,0,0,0,

		1,//ENEMY_MOVE_SPEED,
		DIR_Down,
		0,
		AI_TYPE_Walking_Random,
		0,
		0,
		0,

		true,
		false,
		true,
		true,
	},
	{
		GAME_PIECE_Robot_3,
		
		0,0,0,0,0,0,

		{{98,99,100,101},{98,99,100,101},{98,99,100,101},{98,99,100,101},{98,99,100,101},}, //  dead, up, down, left, right
		{1,1,1,1,1}, //  dead, up, down, left, right
		{4,4,4,4,4}, //  dead, up, down, left, right
		{{0,0}}, // waypoint list populated during init by level data
		
		0,0,0,0,0,0,

		ENEMY_MOVE_SPEED,
		DIR_Right,
		0,
		AI_TYPE_Walking_Tracking,
		0,
		0,
		0,

		true,
		false,
		true,
		true,
	},
	{
		GAME_PIECE_Hatchling,
		
		0,0,0,0,0,0,

		{{73,74,75,74},{73,74,75,74},{73,74,75,74},{73,74,75,74},{73,74,75,74},}, //  dead, up, down, left, right
		{1,1,1,1,1}, //  dead, up, down, left, right
		{4,4,4,4,4}, //  dead, up, down, left, right
		{{0,0}}, // waypoint list populated during init by level data
		
		0,0,0,0,0,0,

		ENEMY_MOVE_SPEED-1,
		DIR_Down,
		0,
		AI_TYPE_None,
		0,
		0,
		0,

		true,
		false,
		true,
		true,
	},
	{
		GAME_PIECE_Wall,
		
		0,0,0,0,0,0,

		{{0,0,0,0},{0,0,0,0},{0,0,0,0},{0,0,0,0},{0,0,0,0}}, // this is a special enemy, uses Jumpan sprite
		{1,1,1,1,1}, //  dead, up, down, left, right
		{1,1,1,1,1}, // this is a special enemy, uses Jumpan sprite
		{{0,0}}, // waypoint list populated during init by level data
		
		0,0,0,0,0,0,

		ENEMY_MOVE_SPEED,
		DIR_None,
		0,
		AI_TYPE_None,
		0,
		0,
		0,

		true,
		false,
		false,
		true,
	},
	{
		GAME_PIECE_Door,
		
		0,0,0,0,0,0,

		{{0,0,0,0},{0,0,0,0},{0,0,0,0},{0,0,0,0},{0,0,0,0}}, // this is a special enemy, uses Jumpan sprite
		{1,1,1,1,1}, //  dead, up, down, left, right
		{1,1,1,1,1}, // this is a special enemy, uses Jumpan sprite
		{{0,0}}, // waypoint list populated during init by level data
		
		0,0,0,0,0,0,

		ENEMY_STOP_SPEED,
		DIR_None,
		0,
		AI_TYPE_None,
		0,
		0,
		0,

		false,
		false,
		false,
		true,
	},
	{
		GAME_PIECE_Transporter,
		
		0,0,0,0,0,0,

		{{76,77,78,79},{76,77,78,79},{76,77,78,79},{76,77,78,79},{76,77,78,79},}, //  dead, up, down, left, right
		{1,1,1,1,1}, //  dead, up, down, left, right
		{4,4,4,4,4}, //  dead, up, down, left, right
		{{0,0}}, // waypoint list populated during init by level data
		
		0,0,0,0,0,0,

		ENEMY_STOP_SPEED,
		DIR_None,
		0,
		AI_TYPE_None,
		0,
		0,
		0,

		true,
		false,
		false,
		true,
	},
	{
		GAME_PIECE_Dragon,
		
		0,0,0,0,0,0,

		{{34,34,35,35},{34,34,35,35},{34,34,35,35},{31,31,33,33},{30,30,32,32},}, //  dead, up, down, left, right
		{1,1,1,1,1}, //  dead, up, down, left, right
		{4,4,4,4,4}, //  dead, up, down, left, right
		{{0,0}}, // waypoint list populated during init by level data
		
		0,0,0,0,0,0,

		ENEMY_MOVE_SPEED,
		DIR_Right,
		0,
		AI_TYPE_Walking_Tracking_Waypoints,
		0,
		CMD_Delay,
		0,

		true,
		false,
		false,
		true,
	},
	{
		GAME_PIECE_Gunslinger,
		
		0,0,0,0,0,0,

		{{0,0,0,0},{2,3,4,5},{2,3,4,5},{7,9,7,9},{6,8,6,8}}, // this is a special enemy, uses Jumpan sprite, but cannot use ropes or chains
		{1,1,1,1,1}, //  dead, up, down, left, right
		{4,4,4,4,4}, // this is a special enemy, uses Jumpan sprite
		{{0,0}}, // waypoint list populated during init by level data
		
		0,0,0,0,0,0,

		ENEMY_MOVE_SPEED,
		DIR_Right,
		0,
		AI_TYPE_Walking_Tracking,
		0,
		0,
		0,

		true,
		false,
		true,
		true,
	},
	{
		GAME_PIECE_Barrel,
		
		0,0,0,0,0,0,

		{{69,69,70,70},{71,71,72,72},{71,71,72,72},{69,69,70,70},{69,69,70,70}}, // this is a special enemy, uses Jumpan sprite, but cannot use ropes or chains
		{1,1,1,1,1}, //  dead, up, down, left, right
		{4,4,4,4,4}, // this is a special enemy, uses Jumpan sprite
		{{0,0}}, // waypoint list populated during init by level data
		
		0,0,0,0,0,0,

		ENEMY_MOVE_SPEED,
		DIR_Right,
		0,
		AI_TYPE_Walking_Waypoints,
		0,
		0,
		0,

		false,
		false,
		false,
		true,
	},
	{
		GAME_PIECE_Freezer,
		
		0,0,0,0,0,0,

		{{73,74,75,74},{73,74,75,74},{73,74,75,74},{73,74,75,74},{73,74,75,74},}, //  dead, up, down, left, right
		{1,1,1,1,1}, //  dead, up, down, left, right
		{4,4,4,4,4}, //  dead, up, down, left, right
		{{0,0}}, // waypoint list populated during init by level data
		
		0,0,0,0,0,0,

		ENEMY_MOVE_SPEED,
		DIR_None,
		0,
		AI_TYPE_None,
		0,
		0,
		0,

		false,
		false,
		false,
		true,
	},
	{
		GAME_PIECE_Falcon,
		
		0,0,0,0,0,0,

		{{40,40,40,40},{40,40,40,40},{40,40,40,40},{37,37,39,39},{36,36,38,38}}, //  dead, up, down, left, right
		{1,1,1,1,1}, //  dead, up, down, left, right
		{4,4,4,4,4}, //  dead, up, down, left, right
		{{0,0}}, // waypoint list populated during init by level data
		
		0,0,0,0,0,0,

		ENEMY_MOVE_SPEED,
		DIR_Right,
		0,
		AI_TYPE_None,
		0,
		0,
		0,

		false,
		false,
		true,
		true,
	},
	{
		GAME_PIECE_Transporter2,
		
		0,0,0,0,0,0,

		{{78,78,78,78},{76,76,77,77},{76,76,77,77},{76,76,77,77},{76,76,77,77}}, //  dead, up, down, left, right
		{1,1,1,1,1}, //  dead, up, down, left, right
		{4,4,4,4,4}, //  dead, up, down, left, right
		{{0,0}}, // waypoint list populated during init by level data
		
		0,0,0,0,0,0,

		ENEMY_MOVE_SPEED,
		DIR_None,
		0,
		AI_TYPE_None,
		0,
		0,
		0,

		true,
		false,
		false,
		true,
	},
};



int2   enemy_count;
uint1  direction;
uint1  previous_waypoint_command = 0;

uint1  move_allowed[5]    = {DIR_None, DIR_None, DIR_None, DIR_None, DIR_None};
bool   move_up_allowed    = false;
bool   move_down_allowed  = false;
bool   move_left_allowed  = false;
bool   move_right_allowed = false;


void DrawEnemies(void)
{
	ENEMY_OBJECT_Type* enemy = &enemy_list[0];
	//uint1 direction;
	int2  width;
	int2 i;
	int2 j;
	uint1* pSprite;

	for (i = 0; i < enemy_count; i++)
	{
		if (enemy->valid == false)
			continue;
			
		if ((enemy->object->flags & FLAG_VISIBLE) && (enemy->delay == 0) && (enemy->waypoint_command != CMD_Delay))
		{
			if ((enemy->enemy_type == GAME_PIECE_Follower))
			{
				MOVE_History_Type* move_history_entry = &move_history[enemy->move_history_index];
				pSprite = (uint1*)j3 + (move_history_entry->sprite_index * (1*8*3));

				enemy->x = move_history_entry->x;
				enemy->y = move_history_entry->y;

				GrayClipSprite8_SMASK_R(enemy->x, enemy->y, 8, pSprite + 8, pSprite + 0, pSprite + 16, PLANE_CurrentPlane0, PLANE_CurrentPlane1);
			}
			else if ((enemy->enemy_type == GAME_PIECE_Wall) || (enemy->enemy_type == GAME_PIECE_Door))
			{
				//GrayUserDrawStr(40,47, "2", 0, 0);
				Sprite8_AND_R(enemy->x, enemy->y, (uint2)enemy->height, door, PLANE_CurrentPlane0);
				Sprite8_AND_R(enemy->x, enemy->y, (uint2)enemy->height, door, PLANE_CurrentPlane1);
			}
			else
			{
				direction = enemy->direction;
				width = enemy->sprite_width[direction];

				if ((enemy->enemy_type == GAME_PIECE_Transporter2) && (collectable_attained >= current_level.activate_required))
					direction = DIR_None;

				pSprite = (uint1*)j3 + (enemy->sprite_index[direction][enemy->anim_step] * 24);

				for (j = 0; j < width; j++)
				{
					GrayClipSprite8_MASK_R(enemy->x + (j * 8), enemy->y, 8, pSprite + 0, pSprite + 8, pSprite + 16, pSprite + 16, PLANE_CurrentPlane0, PLANE_CurrentPlane1);
					pSprite += 24;
				}
			}
		}

		enemy++;
	}
}


void InitEnemies(void)
{
	ENEMY_OBJECT_Type* enemy = &enemy_list[0];
	GAME_PIECE_Type enemy_type = GAME_PIECE_First_Enemy;
	OBJECT_Type* object;
	OBJECT_Type* waypoint_root_object;
	OBJECT_Type* waypoint_list_object;
	bool done = false;
	int2 i;

	enemy_count = 0;
	memset((void*)(enemy_list), 0, sizeof(ENEMY_OBJECT_Type) * ENEMY_COUNT_MAX);

	while ((!done) && (enemy_count < ENEMY_COUNT_MAX))
	{
		object = FindObjectEntryByType(enemy_type, false);

		if (object)
		{
			// initial population of the enemy entry
			memcpy(enemy, &enemy_definition_list[enemy_type - GAME_PIECE_First_Enemy], sizeof(ENEMY_OBJECT_Type));

			// retain a pointer to the enemy object
			enemy->object = object;

			if ((object->flags & FLAG_WAYPOINT) == FLAG_WAYPOINT)
			{
				waypoint_root_object = FindObjectEntryById(object->special_ability, false);
				waypoint_list_object = FindObjectEntryById(waypoint_root_object->waypoint_id, false);

				//enemy->waypoint_index = 0;
				enemy->waypoint_total = waypoint_root_object->count;
				enemy->waypoint_move_step = 0;

				for (i = 0; i < enemy->waypoint_total; i+= 2)
				{
					//enemy->waypoint_type  = waypoint_object->game_piece;
					memcpy(&enemy->waypoint_list_pos[i], &waypoint_list_object->pos[0], 4); // each waypoint data level entry contains 2 position coords
					waypoint_list_object = FindObjectEntryById(waypoint_list_object->id+1, false);
				}
			}
			
			// reset the enemy entry
			ResetEnemy(enemy);

			enemy_count++;
			enemy++;
		}
		else
		{
			enemy_type++;

			if (enemy_type > GAME_PIECE_Last_Enemy)
				done = true;
		}
	}
}


void UpdateEnemies(void)
{
	ENEMY_OBJECT_Type* enemy = &enemy_list[0];
	uint1 speed;
	uint1 direction;
	int2  width;
	int2 i;
	bool new_move_allowed;

	for (i = 0; i < enemy_count; i++)
	{
		if (enemy->valid == false)
			continue;
			
		if (enemy->reset == true)
		{
			ResetEnemy(enemy);
		}

		direction = enemy->direction;
		//direction = DIR_Down;
		width = enemy->sprite_width[direction];
		speed = enemy->speed;

		enemy->anim_step++;
		enemy->direction_count++;
		
		new_move_allowed = false;

		if (enemy->anim_step >= enemy->anim_step_max[direction])
		{
			enemy->anim_step = 0;
		}
		
		if (enemy->direction_count > 30)
			new_move_allowed = true;

/*
		if (enemy->enemy_type != GAME_PIECE_Ufo)
		{
			if (enemy->x > (160-(width * 8)))
				direction = DIR_Left;
			else if (enemy->x <= 0)
				direction = DIR_Right;
			else if (enemy->y > 76)
				direction = DIR_Up;
			else if (enemy->y < 0)
				direction = DIR_Down;
		}
*/
		if (enemy->delay <= 0)
		{
			if (enemy->enemy_type == GAME_PIECE_Falling_Bomb)
			{
				if (direction == DIR_Down)
				{
					enemy->y += speed;

					if (enemy->y >= (88-8))
					{
						direction = DIR_None;

						enemy->direction = direction;
						enemy->x -= 8;
						enemy->anim_step = 0;
						enemy->speed = ENEMY_STOP_SPEED;
					}
				}
				else if (direction == DIR_None)
				{
					if (enemy->anim_step == 0) 
						enemy->reset = true;
				}
			}
			else if (enemy->enemy_type == GAME_PIECE_Ufo)
			{
				if ((enemy->x > 160) || (enemy->x < -(width * 8)) ||
				    (enemy->y >  84) || (enemy->y < -(8)))
				{
					enemy->reset = true;
				}
				else
				{
					if ((direction != DIR_None) && (RandomNumber(50) == 0))
						direction = DIR_Up + RandomNumber(4);

					if (direction == DIR_Up)
						enemy->y -= speed;
					else if (direction == DIR_Down)
						enemy->y += speed;
					else if (direction == DIR_Right)
						enemy->x += speed;
					else if (direction == DIR_Left)
						enemy->x -= speed;
					else if (direction == DIR_None)
						enemy->y += (speed + 1);
				}
			}
			else if (enemy->enemy_type == GAME_PIECE_Bat)
			{
				if (enemy->hold == false)
				{
					if ((RandomNumber(20) == 0))
					{
						int2 xdiff = enemy->x - jumpman.x;
						int2 ydiff = enemy->y - jumpman.y;

						if (abs(xdiff) < abs(ydiff))
						{
							if (enemy->y < jumpman.y)
								direction = DIR_Down;
							else
								direction = DIR_Up;
						}
						else //if (abs(ydiff) < 8)
						{
							if (enemy->x < jumpman.x)
								direction = DIR_Right;
							else
								direction = DIR_Left;
						}
					}

					if (enemy->anim_step & 1)
					{
						if (direction == DIR_Up)
							enemy->y -= speed;
						else if (direction == DIR_Down)
							enemy->y += speed;
						else if (direction == DIR_Right)
							enemy->x += speed;
						else if (direction == DIR_Left)
							enemy->x -= speed;
					}
				}
				else
				{
					if (collectable_trigger == true)
					{
						collectable_trigger = false;
						enemy->hold = false;
					}
				}
			}
			else if (enemy->enemy_type == GAME_PIECE_Freezer)
			{
				if (enemy->hold == false)
				{
					int2 xdiff = enemy->x - jumpman.x;
					int2 ydiff = enemy->y - jumpman.y;

					if ((RandomNumber(5) == 0) || (jumpman.freeze_count))
					{
						if (abs(xdiff) < abs(ydiff))
						{
							if (enemy->y < jumpman.y)
								direction = DIR_Down;
							else
								direction = DIR_Up;
						}
						else //if (abs(ydiff) < 8)
						{
							if (enemy->x < jumpman.x)
								direction = DIR_Right;
							else
								direction = DIR_Left;
						}
					}
					
					if (jumpman.freeze_count)
					{
						if (direction == DIR_Down)
							direction = DIR_Up;
						else if (direction == DIR_Up)
							direction = DIR_Down;
						else if (direction == DIR_Left)
							direction = DIR_Right;
						else if (direction == DIR_Right)
							direction = DIR_Left;
					}

					if (direction == DIR_Up)
					{
						if ((enemy->y - speed) > -6)
							enemy->y -= speed;
					}
					else if (direction == DIR_Down)
					{
						if ((enemy->y + speed) < 86)
							enemy->y += speed;
					}
					else if (direction == DIR_Left)
					{
						if ((enemy->x - speed) > -6)
							enemy->x -= speed;
					}
					else if (direction == DIR_Right)
					{
						if ((enemy->x + speed) < 158)
							enemy->x += speed;
					}
				}
			}
			else if (enemy->enemy_type == GAME_PIECE_Bird)
			{
				enemy->counter--;

				if (enemy->counter <= 0)
				{
					int2 random_num = RandomNumber(3);

					if (random_num == 0)
						direction = DIR_Down;
					else if (random_num == 1)
						direction = DIR_Up;
					else
						direction = DIR_None;

					enemy->counter = RandomNumber(50);
				}

				if (enemy->direction == DIR_Down)
					enemy->y++;
				else if (enemy->direction == DIR_Up)
					enemy->y--;

				enemy->x += current_level.wind_speed;

				if (enemy->x > 156)
				{
					enemy->x = -3;
					enemy->delay = RandomNumber(50);
				}
			}
			else if (enemy->enemy_type == GAME_PIECE_Falcon)
			{
				int2   x_diff = enemy->x - jumpman.x;
				int2   y_diff = enemy->y - jumpman.y;
/*				
				enemy->counter--;

				if (enemy->counter <= 0)
				{
					int2 random_num = RandomNumber(3);

					if (random_num == 0)
						direction = DIR_Down;
					else if (random_num == 1)
						direction = DIR_Up;
					else
						direction = DIR_None;

					enemy->counter = RandomNumber(50);
				}
*/				
				if ((y_diff <= 10) && (y_diff >= 0))
				{
					enemy->y += RandomNumber(2);	
									
					if (x_diff > 0)
						direction = DIR_Left;
					else if (x_diff < 0)
						direction = DIR_Right;
					else
						direction = DIR_Up;
				}
				else if (RandomNumber(4) == 0)
				{
						direction = DIR_Up;
				}
				else if (RandomNumber(20) == 0)
				{
					switch (RandomNumber(2))
					{
						case 0: 
							direction = DIR_Left;
							break;
						case 1: 
							direction = DIR_Right;
							break;
					}
				}

				enemy->y--;					

				if (direction == DIR_Left)
				{
					enemy->x -= 4;
					//enemy->y -= 2;
				}
				else if (direction == DIR_Right)
				{
					enemy->x += 4;
					//enemy->y -= 2;
				}
					
				if (enemy->y < -4)
				{
					enemy->y = 84;
					enemy->delay = RandomNumber(10);
				}

				if (enemy->x < -3)
				{
					enemy->x = 160;
					enemy->delay = RandomNumber(10);
				}
				else if (enemy->x > 160)
				{
					enemy->x = -3;
					enemy->delay = RandomNumber(10);
				}
			}
			else if (enemy->enemy_type == GAME_PIECE_Follower)
			{
				if ((enemy->hold == false) && (jumpman.action <= ACTION_Rope))
				{
					enemy->move_history_index++;
					enemy->move_history_index &= MOVE_HISTORY_MASK;
				}
			}
			else if (enemy->enemy_type == GAME_PIECE_Wall)
			{
				if (enemy->hold == false)
				{
					if (direction == DIR_Left)
					{
						enemy->x -= ENEMY_MOVE_SPEED;

						if (enemy->x < 0)
							enemy->x = 156;
					}
					else if (direction == DIR_Right)
					{
						enemy->x += ENEMY_MOVE_SPEED;

						if (enemy->x >= 156)
							enemy->x = 0;
					}
				}
			}
			else if (enemy->enemy_type == GAME_PIECE_Door)
			{
				if (enemy->hold == false)
				{
					if (direction == DIR_Up)
					{
				//GrayUserDrawStr(40,47, "UP", 0, 0);
						enemy->height -= ENEMY_MOVE_SPEED;

						if (enemy->height < 2)
						{
							enemy->height = 2;
							
							if (enemy->object->flags & FLAG_TOGGLE)
								direction = DIR_Down;
							//else
							//	enemy->hold = true;
						}
					}
					else if (direction == DIR_Down)
					{
				//GrayUserDrawStr(40,57, "DOWN", 0, 0);
						enemy->height += ENEMY_MOVE_SPEED;

						if (enemy->height > enemy->height_max)
						{
							enemy->height = enemy->height_max;
							
							if (enemy->object->flags & FLAG_TOGGLE)
								direction = DIR_Up;
							//else
							//	enemy->hold = true;
						}
					}
				}
			}
			else if (enemy->enemy_type == GAME_PIECE_Robot_1)
			{
				//sprintf(tmpStr1, "xy: (%d, %d)", enemy->x, enemy->y);
				//GrayUserDrawStr(40,50, tmpStr1, 0, 0);
				//GrayUserDrawStr(40,57, tmpStr2, 0, 0);

				if (enemy->hold == false)
				{
					if (enemy->waypoint_move_step > 0)
					{
						//uint1 enemy_bkgd[2];
						enemy->waypoint_move_step -= 2;
						//GrayUserDrawStr(74,1, "MOVING TO WP", 0, 0);

						if (direction == DIR_Down)
							enemy->y += speed;
						else if (direction == DIR_Up)
							enemy->y -= speed;
						else if (direction == DIR_Left)
							enemy->x -= speed;
						else if (direction == DIR_Right)
							enemy->x += speed;
						
						VerticalAdjustEnemy(enemy);

						//if ((direction == DIR_Left) || (direction == DIR_Right))
						//{
						//	ClipSprite8Get_R(enemy->x, enemy->y+7, 2, plane_active_1, enemy_bkgd);
						//		
						//	if (enemy_bkgd[0] & 0x3c)
						//		enemy->y--;
						//	else if ((enemy_bkgd[1] & 0x3c) == 0)
						//		enemy->y++;
						//}

					}
					else
					{
						// get the next waypoint
						uvect1* waypoint;

						enemy->waypoint_index++;
						
						if (enemy->waypoint_index >= enemy->waypoint_total)
							enemy->waypoint_index = 0;

						waypoint = &enemy->waypoint_list_vec[enemy->waypoint_index];

						//	sprintf(tmpStr2, "%u: (%d, %d)", enemy->waypoint_index, waypoint->direction, waypoint->distance);
						//	GrayUserDrawStr(40,50, tmpStr2, 0, 0);
							
						direction = waypoint->direction;

						if (direction == DIR_None)
						{
							enemy->hold = true;
							enemy->object->flags |= FLAG_HOLD;
							//GrayUserDrawStr(74,8, "END WP", 0, 0);
						}
						else
						{

							enemy->waypoint_move_step = waypoint->distance;

							if ((direction == DIR_Left) || (direction == DIR_Right))
								enemy->waypoint_move_step *= 4;

							//GrayUserDrawStr(74,15, "NEXT WP", 0, 0);
						}
					}
				}
				else
				{
					if ((enemy->object->flags & FLAG_HOLD) == 0)
					{
						enemy->waypoint_move_step = 0;
						enemy->hold = false;

						//GrayUserDrawStr(74,22, "RELEASE", 0, 0);
					}
				}
			}
			else if ((enemy->enemy_type == GAME_PIECE_Dragon) || (enemy->enemy_type == GAME_PIECE_Barrel))
			{
				if (enemy->waypoint_move_step == 0)
				{
					// get the next waypoint
					uvect1* waypoint;
					
					waypoint = &enemy->waypoint_list_vec[enemy->waypoint_index];

					if (waypoint->direction <= DIR_Right)
					{
						direction = waypoint->direction;
						enemy->direction = direction;

						enemy->waypoint_move_step = waypoint->distance;

						if ((direction == DIR_Left) || (direction == DIR_Right))
							enemy->waypoint_move_step *= 4;
							
						enemy->waypoint_index++;
							
						if (enemy->waypoint_index >= enemy->waypoint_total)
							enemy->waypoint_index = 0;
							
											
						enemy->waypoint_command = direction;

					}
					else
					{
						uint1 command = waypoint->direction;
						uint1 value = waypoint->distance;
						
						if (command == CMD_Delay)
						{
							enemy->delay = value;// + RandomNumber(10);
							enemy->waypoint_move_step = 0;
							enemy->waypoint_index++;
							
							if (enemy->waypoint_index >= enemy->waypoint_total)
								enemy->waypoint_index = 0;
						}
						else if (command == CMD_Goto)
						{
							enemy->waypoint_index = value;
							enemy->waypoint_move_step = 0;	
							
							//if (enemy->prev_waypoint_command == CMD_Delay)
							//{
							//	enemy->delay = 2;
							//}
						}
						else if (command == CMD_Cmp_YLT)
						{
							if (enemy->y < jumpman.y)
							{
								enemy->waypoint_index = value;
								enemy->waypoint_move_step = 0;					
							}
							else
							{
								enemy->waypoint_index++;
									
								if (enemy->waypoint_index >= enemy->waypoint_total)
									enemy->waypoint_index = 0;
							}
						}
						
						enemy->waypoint_command = command;
					}
				}	
				else // if (enemy->waypoint_move_step > 0)
				{
					enemy->waypoint_move_step -= 2;
					
					if (enemy->waypoint_move_step < 0)
						enemy->waypoint_move_step = 0;

					if (direction == DIR_Down)
						enemy->y += speed;
					else if (direction == DIR_Up)
						enemy->y -= speed;
					else if (direction == DIR_Left)
						enemy->x -= speed;
					else if (direction == DIR_Right)
						enemy->x += speed;
					
					if (enemy->enemy_type == GAME_PIECE_Barrel)
						VerticalAdjustEnemy(enemy);
				}
			}
			else if ((enemy->enemy_type == GAME_PIECE_Robot_2))
			{
				int2   n = 0;
				uint1  temp_direction = DIR_None;
				int2   continue_move_chance;
				int2   fifty_fifty_chance;
				
				continue_move_chance = RandomNumber(100);
				fifty_fifty_chance = RandomNumber(2);

#if 1
				VerticalAdjustEnemy(enemy);
				CheckMoveEnemy(enemy);
				

				if (direction == DIR_Down)
				{
					if (move_left_allowed && move_right_allowed && new_move_allowed)
						temp_direction = DIR_Left + fifty_fifty_chance;
					else if (move_left_allowed && fifty_fifty_chance && new_move_allowed)
						temp_direction = DIR_Left;
					else if (move_right_allowed && fifty_fifty_chance && new_move_allowed)
						temp_direction = DIR_Right;
					else if (move_down_allowed && continue_move_chance)
						temp_direction = DIR_Down;
					else if (move_up_allowed && fifty_fifty_chance)
						temp_direction = DIR_Up;
				}
				else if (direction == DIR_Up)
				{
					if (move_left_allowed && move_right_allowed && new_move_allowed)
						temp_direction = DIR_Left + fifty_fifty_chance;
					else if (move_left_allowed && fifty_fifty_chance && new_move_allowed)
						temp_direction = DIR_Left;
					else if (move_right_allowed && fifty_fifty_chance && new_move_allowed)
						temp_direction = DIR_Right;
					else if (move_up_allowed && continue_move_chance)
						temp_direction = DIR_Up;
					else if (move_down_allowed && fifty_fifty_chance)
						temp_direction = DIR_Down;
				}
				else if (direction == DIR_Left)
				{
					if (move_up_allowed && move_down_allowed && new_move_allowed)
						temp_direction = DIR_Up + fifty_fifty_chance;
					else if (move_down_allowed && fifty_fifty_chance && new_move_allowed)
						temp_direction = DIR_Down;
					else if (move_up_allowed && fifty_fifty_chance && new_move_allowed)
						temp_direction = DIR_Up;
					else if (move_left_allowed && continue_move_chance)
						temp_direction = DIR_Left;
					else if (move_right_allowed && fifty_fifty_chance)
						temp_direction = DIR_Right;
				}
				else if (direction == DIR_Right)
				{
					if (move_up_allowed && move_down_allowed && new_move_allowed)
						temp_direction = DIR_Up + fifty_fifty_chance;
					else if (move_down_allowed && fifty_fifty_chance && new_move_allowed)
						temp_direction = DIR_Down;
					else if (move_up_allowed && fifty_fifty_chance && new_move_allowed)
						temp_direction = DIR_Up;
					else if (move_right_allowed && continue_move_chance)
						temp_direction = DIR_Right;
					else if (move_left_allowed && fifty_fifty_chance)
						temp_direction = DIR_Left;
				}
#if 1
				if (temp_direction == DIR_None)
				{
					// Let's loop a max of 32 times to find a new direction
					while (n++ < 32)
					{
						temp_direction = move_allowed[RandomNumber(4) + 1];
						if (temp_direction != DIR_None)
							break;
					}
				}
#endif
#if 0
if (enemy == &enemy_list[0])
{
	char buffer[32];
	sprintf(buffer, "%02d:%02x,%02x,%02x,%02x,", enemy->direction,temp_direction,fifty_fifty_chance,continue_move_chance,new_move_allowed);
	GrayUserDrawStr(40,32, buffer, 0, 0);
	sprintf(buffer, "%02d:%02x,%02x,%02x,%02x,", enemy->y,move_up_allowed,move_down_allowed,move_left_allowed,move_right_allowed);
	GrayUserDrawStr(40,40, buffer, 0, 0);
}
#endif
				

				if (direction != temp_direction)
					enemy->direction_count = 0;
					
				direction = temp_direction;
					
#endif
				//direction = GetAIType3Move(enemy);

				if (direction == DIR_Down)
					//enemy->y += (speed - 1);
					enemy->y++;
				else if (direction == DIR_Up)
					//enemy->y -= (speed - 1);
					enemy->y--;
				else if (direction == DIR_Left)
					//enemy->x -= speed;
					enemy->x -= 1;
				else if (direction == DIR_Right)
					//enemy->x += speed;
					enemy->x += 1;
					
				if (enemy->object->special_ability == SPECIAL_ABILITY_Gunfire)
				{
					int2 x_diff = enemy->x - jumpman.x;
					int2 y_diff = enemy->y - jumpman.y;
					int2 x_y_diff = (EXT_SHORTABS(x_diff) - EXT_SHORTABS(y_diff)) >> 2;
					int2 bullet_direction = DIR_None;
					
					if (x_diff == 0)
					{
						if (y_diff > 0)
							bullet_direction = DIR_Up;
						else if (y_diff < 0)
							bullet_direction = DIR_Down;
					}
					else if (y_diff == 0)
					{
						if (x_diff > 0)
							bullet_direction = DIR_Left;
						else if (x_diff < 0)
							bullet_direction = DIR_Right;
					}
					else if (x_y_diff == 0)
					{
						if ((x_diff > 0) && (y_diff > 0))
							bullet_direction = DIR_UpLeft;
						if ((x_diff > 0) && (y_diff < 0))
							bullet_direction = DIR_DownLeft;
						if ((x_diff < 0) && (y_diff > 0))
							bullet_direction = DIR_UpRight;
						if ((x_diff < 0) && (y_diff < 0))
							bullet_direction = DIR_DownRight;
					}
					
					if (enemy->bullet_fire_count == 0)
					{
						enemy->bullet_fire_count = 8;
						FireBullet(enemy, GAME_PIECE_Enemy_Gunfire, bullet_direction);
					}
				}				
			}
			else if ((enemy->enemy_type == GAME_PIECE_Robot_3) || 
							 (enemy->enemy_type == GAME_PIECE_Gunslinger))
			{
				int2   x_diff = enemy->x - jumpman.x;
				int2   y_diff = enemy->y - jumpman.y;
#if 1
				int2   n = 0;
				uint1  temp_direction = DIR_None;
				uint1  preferred_x_move = DIR_None;
				uint1  preferred_y_move = DIR_None;
				int2   continue_move_chance = RandomNumber(64);
				
				if (x_diff < 0)
					preferred_x_move = DIR_Right;
				else if (x_diff > 0)
					preferred_x_move = DIR_Left;
				else
					preferred_x_move = DIR_None;

				if (y_diff < 0)
					preferred_y_move = DIR_Down;
				else if (y_diff > 0)
					preferred_y_move = DIR_Up;
				else
					preferred_y_move = DIR_None;

				VerticalAdjustEnemy(enemy);
				CheckMoveEnemy(enemy);
				
				if (enemy->direction_count > 5)
					new_move_allowed = true;
				else
					new_move_allowed = false;


				if (direction == DIR_Up)
				{
					if ((preferred_x_move == DIR_Right) && (move_right_allowed) && (new_move_allowed))
						temp_direction = DIR_Right;
					else if ((preferred_x_move == DIR_Left) && (move_left_allowed) && (new_move_allowed))
						temp_direction = DIR_Left;
					else if (move_up_allowed && continue_move_chance)
						temp_direction = DIR_Up;
					else if (move_down_allowed)
						temp_direction = DIR_Down;
				}
				else if (direction == DIR_Down)
				{
					if ((preferred_x_move == DIR_Right) && (move_right_allowed) && (new_move_allowed))
						temp_direction = DIR_Right;
					else if ((preferred_x_move == DIR_Left) && (move_left_allowed) && (new_move_allowed))
						temp_direction = DIR_Left;
					else if (move_down_allowed && continue_move_chance)
						temp_direction = DIR_Down;
					else if (move_up_allowed)
						temp_direction = DIR_Up;
				}
				else if (direction == DIR_Left)
				{
					if ((preferred_y_move == DIR_Up) && (move_up_allowed) && (new_move_allowed))
						temp_direction = DIR_Up;
					else if ((preferred_y_move == DIR_Down) && (move_down_allowed) && (new_move_allowed))
						temp_direction = DIR_Down;
					else if (move_left_allowed && continue_move_chance)
						temp_direction = DIR_Left;
					else if (move_right_allowed)
						temp_direction = DIR_Right;
				}
				else if (direction == DIR_Right)
				{
					if ((preferred_y_move == DIR_Up) && (move_up_allowed) && (new_move_allowed))
						temp_direction = DIR_Up;
					else if ((preferred_y_move == DIR_Down) && (move_down_allowed) && (new_move_allowed))
						temp_direction = DIR_Down;
					else if (move_right_allowed && continue_move_chance)
						temp_direction = DIR_Right;
					else if (move_left_allowed)
						temp_direction = DIR_Left;
				}
#if 1
				if (temp_direction == DIR_None)
				{
					// Let's loop a max of 32 times to find a new direction
					while (n++ < 32)
					{
						temp_direction = move_allowed[RandomNumber(4) + 1];
						if (temp_direction != DIR_None)
							break;
					}
				}
#endif
				
				if (direction != temp_direction)
					enemy->direction_count = 0;
					
				direction = temp_direction;
#endif


				
				//direction = GetAIType3Move(enemy);


				if (direction == DIR_Down)
					//enemy->y += (speed - 1);
					enemy->y++;
					//enemy->y += speed;
				else if (direction == DIR_Up)
					//enemy->y -= (speed - 1);
					enemy->y--;
					//enemy->y -= speed;
				else if (direction == DIR_Left)
					enemy->x -= speed;
				else if (direction == DIR_Right)
					enemy->x += speed;
					
				if (enemy->object->special_ability == SPECIAL_ABILITY_Gunfire)
				{
					int2 x_y_diff = (EXT_SHORTABS(x_diff) - EXT_SHORTABS(y_diff)) >> 2;
					int2 bullet_direction = DIR_None;
					
					if (x_diff == 0)
					{
						if (y_diff > 0)
							bullet_direction = DIR_Up;
						else if (y_diff < 0)
							bullet_direction = DIR_Down;
					}
					else if (y_diff == 0)
					{
						if (x_diff > 0)
							bullet_direction = DIR_Left;
						else if (x_diff < 0)
							bullet_direction = DIR_Right;
					}
					else if (x_y_diff == 0)
					{
						if ((x_diff > 0) && (y_diff > 0))
							bullet_direction = DIR_UpLeft;
						if ((x_diff > 0) && (y_diff < 0))
							bullet_direction = DIR_DownLeft;
						if ((x_diff < 0) && (y_diff > 0))
							bullet_direction = DIR_UpRight;
						if ((x_diff < 0) && (y_diff < 0))
							bullet_direction = DIR_DownRight;
					}
					
					if (enemy->bullet_fire_count == 0)
					{
						enemy->bullet_fire_count = 8;
						FireBullet(enemy, GAME_PIECE_Enemy_Gunfire, bullet_direction);
					}
				}			
			}
			else if (enemy->enemy_type == GAME_PIECE_Hatchling)
			{
				if (enemy->object->flags & FLAG_VISIBLE)
				{
					if (enemy->hold == false)
					{
						int2 xdiff = enemy->x - jumpman.x;
						int2 ydiff = enemy->y - jumpman.y;
	
						if (ydiff < 0)
							enemy->y += 1;
						else if (ydiff > 0)
							enemy->y -= 1;
	
						if (xdiff < 0)
							enemy->x += 1;
						else if (xdiff > 0)
							enemy->x -= 1;
					}
				}
			}
			else if (enemy->enemy_type == GAME_PIECE_Transporter)
			{
				enemy->counter++;

				if (enemy->counter > 30)
				{
					enemy->counter = 0;

					enemy->waypoint_index++;

					if (enemy->waypoint_index >= enemy->waypoint_total)
						enemy->waypoint_index = 0;

					enemy->x = enemy->waypoint_list_pos[enemy->waypoint_index].x * 4;
					enemy->y = enemy->waypoint_list_pos[enemy->waypoint_index].y;
				}
			}
			else if (enemy->enemy_type == GAME_PIECE_Transporter2)
			{
				if (enemy->waypoint_move_step <= 0)
				{
					// get the next waypoint
					uvect1* waypoint;
					
					waypoint = &enemy->waypoint_list_vec[enemy->waypoint_index];

				//GrayUserDrawStr(74,8, "WP1", 0, 0);
				
					if (waypoint->direction <= DIR_Right)
					{
						direction = waypoint->direction;
						enemy->direction = direction;

						enemy->waypoint_move_step = waypoint->distance;

						if ((direction == DIR_Left) || (direction == DIR_Right))
							enemy->waypoint_move_step *= 4;
							
						enemy->waypoint_index++;
							
						if (enemy->waypoint_index >= enemy->waypoint_total)
							enemy->waypoint_index = 0;
					}
					else
					{
				//GrayUserDrawStr(74,8, "WP2", 0, 0);
						uint1 command = waypoint->direction;
						uint1 value = waypoint->distance;
						
						if (command == CMD_Goto)
						{
							enemy->waypoint_index = value;
							enemy->waypoint_move_step = 0;	
						}
					}
				}	
				else // if (enemy->waypoint_move_step > 0)
				{
					enemy->waypoint_move_step -= speed;
					
					if (enemy->waypoint_move_step < 0)
						enemy->waypoint_move_step = 0;

					if (direction == DIR_Down)
						enemy->y += speed;
					else if (direction == DIR_Up)
						enemy->y -= speed;
					else if (direction == DIR_Left)
						enemy->x -= speed;
					else if (direction == DIR_Right)
						enemy->x += speed;
				}
			}
		}
		else
		{
			enemy->delay--;
		}

		// update the enemy's direction
		enemy->direction = direction;
		
		if (enemy->bullet_fire_count)
			enemy->bullet_fire_count--;

		// move to the next enemy
		enemy++;
	}
}

void VerticalAdjustEnemy(ENEMY_OBJECT_Type* enemy)
{
	uint1  enemy_bkgd[2];

	if ((enemy->direction == DIR_Left) || (enemy->direction == DIR_Right))
	{
		ClipSprite8Get_R(enemy->x, enemy->y+7, 2, plane_active_1, enemy_bkgd);

		if (enemy_bkgd[0] & 0x3c)
			enemy->y--;
		else if ((enemy_bkgd[1] & 0x3c) == 0)
			enemy->y++;
	}
}

void CheckMoveEnemy(ENEMY_OBJECT_Type* enemy)
{
	if (enemy->enemy_type == GAME_PIECE_Freezer)
	{
		move_allowed[DIR_Up] = DIR_Up;
		move_up_allowed = true;

		move_allowed[DIR_Down] = DIR_Down;
		move_down_allowed = true;

		move_allowed[DIR_Left] = DIR_Left;
		move_left_allowed = true;

		move_allowed[DIR_Right] = DIR_Right;
		move_right_allowed = true;
	}
	else
	{
		uint1  enemy_bkgd[8];
		//uint1* p_enemy_bkgd = enemy_bkgd;
	
		move_allowed[DIR_Up]    = DIR_None;
		move_allowed[DIR_Down]  = DIR_None;
		move_allowed[DIR_Left]  = DIR_None;
		move_allowed[DIR_Right] = DIR_None;
		
		move_up_allowed    = false;
		move_down_allowed  = false;
		move_left_allowed  = false;
		move_right_allowed = false;
	
		ClipSprite8Get_R(enemy->x, enemy->y+4, 8, plane_active_0, enemy_bkgd);

#if 0
if (enemy == &enemy_list[0])
{
	char buffer[32];
	sprintf(buffer, "%02d:%02x,%02x,%02x,%02x,", enemy->x,enemy_bkgd[4],enemy_bkgd[5],enemy_bkgd[6],enemy_bkgd[7]);
	GrayUserDrawStr(40,47, buffer, 0, 0);
}
#endif
	
		if ((enemy_bkgd[0] == 0xff) ||
				(enemy_bkgd[1] == 0xff) ||
				(enemy_bkgd[2] == 0xff) ||
				(enemy_bkgd[3] == 0xff))
		{
			move_allowed[DIR_Up] = DIR_Up;
			move_up_allowed = true;
		}
	
		if ((enemy_bkgd[4] == 0xff) ||
				(enemy_bkgd[5] == 0xff) ||
				(enemy_bkgd[6] == 0xff) ||
				(enemy_bkgd[7] == 0xff))
		{
			move_allowed[DIR_Down] = DIR_Down;
			move_down_allowed = true;
		}
	
		ClipSprite8Get_R(enemy->x, enemy->y+7, 3, plane_active_1, enemy_bkgd);
	
		if (((enemy_bkgd[0] & 0xf0)) ||
				((enemy_bkgd[1] & 0xf0)) ||
				((enemy_bkgd[2] & 0xf0)))
		{
			move_allowed[DIR_Left] = DIR_Left;
			move_left_allowed = true;
		}
		
		if (((enemy_bkgd[0] & 0x0f)) ||
				((enemy_bkgd[1] & 0x0f)) ||
				((enemy_bkgd[2] & 0x0f)))
		{
			move_allowed[DIR_Right] = DIR_Right;
			move_right_allowed = true;
		}
		
#if 0
if (enemy == &enemy_list[0])
{
	char buffer[32];
	uint1 moves = (move_right_allowed << 3) | (move_left_allowed << 2) | (move_down_allowed << 1) | (move_up_allowed);
	sprintf(buffer, "%02x:%02x,%02x,%02x,%02x,%02x,%02x", enemy->x,moves,enemy->direction, enemy->direction_count,enemy_bkgd[0],enemy_bkgd[1],enemy_bkgd[2]);
	GrayUserDrawStr(40,47, buffer, 0, 0);
}
#endif
		
	}
}

		
void CheckEnemies(void)
{
	uint1 flag_test_value = (FLAG_ACTIVE | FLAG_VISIBLE);
	int2 i;
	int2 j;

	ENEMY_OBJECT_Type* enemy = &enemy_list[0];

	for (i = 0; i < enemy_count; i++)
	{
		uint1 test_value = 0;

		if (enemy->valid == false)
			continue;
			
		if (enemy->delay == 0)
		{
			if (enemy->object->game_piece == GAME_PIECE_Wall)
			{
				//BOUNDS_COLLIDE2HW(x0, y0, x1, y1, w0, w1, h0, h1)
				//test_value = BOUNDS_COLLIDE2HW(enemy->x + 2, enemy->y + 1, jumpman.x_orig + 1, jumpman.y_orig + 1, 4, 6, enemy->height-2, 6);
				test_value |= BOUNDS_COLLIDE_UTIL(enemy->x + 2, enemy->y + 1, jumpman.x_orig + 1, jumpman.y_orig + 1, 6, 6, enemy->height-2, 6);
				
				if (test_value)
				{
					if (enemy->direction == DIR_Left)
						jumpman.x -= enemy->speed;
					else if (enemy->direction == DIR_Right)
						jumpman.x += enemy->speed;
				}
			}
			else if (enemy->object->game_piece == GAME_PIECE_Door)
			{
				//BOUNDS_COLLIDE2HW(x0, y0, x1, y1, w0, w1, h0, h1)
				test_value = BOUNDS_COLLIDE_UTIL(enemy->x, enemy->y, jumpman.x_orig, jumpman.y_orig, 4, 8, enemy->height, 8);
				
				if (test_value)
				{
					if (jumpman.action < ACTION_Fall)
						SetJumpmanAction(ACTION_Fall, DIR_Down);
				}
			}
			else if (enemy->object->game_piece != GAME_PIECE_Transporter)
			{
				if ((enemy->object->flags & flag_test_value) == flag_test_value)
				{
					for (j = 0; j < enemy->sprite_width[enemy->direction]; j++)
					{
						//test_value |= BOUNDS_COLLIDE82H(enemy->x + (j * 8), enemy->y + 1, jumpman.x_orig, jumpman.y_orig + 1, 6, 6);
						test_value |= BOUNDS_COLLIDE_UTIL(enemy->x + (j * 8) + 1, enemy->y + 1, jumpman.x_orig + 1, jumpman.y_orig + 1, 6, 6, 6, 6);
					}
					/*
					uint1  test_value = PixCollide8_R(
																	(int2)collectable->x+4,
																	(int2)collectable->y+2,
																	(int2)jumpman.x_orig,
																	(int2)jumpman.y_orig,
																	(uint2)10,
																	jumpman_test_sprite);
					*/

					if (test_value)
					{
						//GrayUserDrawStr(74,1, "Collision", 0, 0);
						if (jumpman.action < ACTION_Fall)
						{
							if (enemy->object->game_piece == GAME_PIECE_Freezer)
							{
								jumpman.freeze_count = 20;
							}
							else if ((enemy->enemy_type == GAME_PIECE_Transporter2) && (collectable_attained >= current_level.activate_required))
							{
								ExecuteScript(current_level.activate_script_id);
								enemy->valid = false;
							}
							else
							{
								SetJumpmanAction(ACTION_Fall, DIR_Down);
							}
						}

						break;
					}
				}
			}
		}

		enemy++;
	}
}

void ResetEnemy(ENEMY_OBJECT_Type* enemy)
{
	uint1 enemy_type = enemy->enemy_type;
	
	enemy->anim_step = 0;
	enemy->direction_count = 0;
	//enemy->speed     = ENEMY_STOP_SPEED;
	//enemy->speed     = 2;
	//enemy->direction = DIR_Down;
	enemy->direction = DIR_None;
	enemy->delay     = 0;
	enemy->counter   = 0;
	enemy->x         = enemy->object->x * 4;
	enemy->y         = enemy->object->y;
	enemy->waypoint_index      = 0;
	enemy->waypoint_move_step  = 0;
	enemy->reset     = false;
	enemy->valid     = true;	
	enemy->bullet_fire_count = 0;

	if (enemy_type == GAME_PIECE_Falling_Bomb)
	{
		enemy->x         = RandomNumber(152);
		enemy->y         = -8;
		enemy->direction = DIR_Down;
		enemy->speed     = ENEMY_MOVE_SPEED;
		enemy->delay     = RandomNumber(100);
	}
	else if (enemy_type == GAME_PIECE_Ufo)
	{
		uint1 direction = DIR_Up + RandomNumber(4);
		int2  x = RandomNumber(152);
		int2  y = RandomNumber(82);

		if (direction == DIR_Up)
			y = 88;
		else if (direction == DIR_Down)
			y = -8;
		else if (direction == DIR_Right)
			x = -16;
		else // if (direction == DIR_Left)
			x = 160;

		enemy->x         = x;
		enemy->y         = y;
		enemy->direction = direction;
		enemy->speed     = ENEMY_MOVE_SPEED;
		enemy->object->flags     = FLAG_VA;
		enemy->delay     = RandomNumber(100);
	}
	else if (enemy_type == GAME_PIECE_Bat)
	{
		enemy->speed     = ENEMY_MOVE_SPEED-1;
		enemy->direction = DIR_None;
		enemy->delay     = 0;
	}
	else if (enemy_type == GAME_PIECE_Bird)
	{
		enemy->speed     = ENEMY_MOVE_SPEED-1;
		enemy->direction = DIR_None;
		enemy->delay     = RandomNumber(50);
		enemy->counter   = RandomNumber(50);
	}
	else if (enemy_type == GAME_PIECE_Follower)
	{
		enemy->object->flags &= ~(FLAG_VISIBLE | FLAG_ACTIVE);
		enemy->hold = true;
		enemy->delay = 8;
		enemy->move_history_index = 0;
	}
	else if (enemy_type == GAME_PIECE_Robot_1)
	{
		enemy->hold = true;
		enemy->direction = DIR_Down;
		enemy->delay = 0;
		enemy->waypoint_index = -1;
		enemy->speed = 2;
	}
	else if (enemy_type == GAME_PIECE_Hatchling)
	{
		//enemy->hold = true;
		enemy->object->flags = FLAG_NONE;
	}
	else if (enemy_type == GAME_PIECE_Barrel)
	{
		enemy->speed     = 2;
		enemy->delay     = 0;
	}
	else if ((enemy_type == GAME_PIECE_Wall) ||  (enemy_type == GAME_PIECE_Door))
	{	
		enemy->direction  = enemy->object->height_and_dir & 0x07;
		enemy->height_max = enemy->object->height_and_dir >> 3;
		enemy->height     = enemy->height_max;
		
		if (enemy->direction > DIR_Right)
			enemy->direction = DIR_None;
			
		if (enemy_type == GAME_PIECE_Door)
		{
			if (enemy->direction == DIR_Up)
				enemy->height = 2;
		}
	}
	else if (enemy_type == GAME_PIECE_Dragon)
	{
		enemy->waypoint_command = CMD_Delay;
	}
	else if (enemy_type != GAME_PIECE_Fire)
	{
		enemy->delay = 30;
	}
}

void ExtinguishAllFires(void)
{
	ENEMY_OBJECT_Type* enemy = &enemy_list[0];
	int2 i;

	for (i = 0; i < enemy_count; i++)
	{
		if (enemy->enemy_type == GAME_PIECE_Fire)
		{
			enemy->object->flags = FLAG_NONE;
		}

		enemy++;
	}
}

void ReleaseNextEnemy(uint1 enemy_type)
{
	ENEMY_OBJECT_Type* enemy = &enemy_list[0];
	int2 i;

	for (i = 0; i < enemy_count; i++)
	{
		if (enemy->enemy_type == enemy_type)
		{
			if (enemy->hold == true)
			{
				enemy->object->flags |= (FLAG_VISIBLE | FLAG_ACTIVE);
				//enemy->object->flags = (FLAG_VISIBLE | FLAG_ACTIVE);
				enemy->hold = false;
				enemy->delay = 20;
				enemy->move_history_index = 0;;

				break;
			}
		}

		enemy++;
	}
}

void ReleaseHatchling(int2 enemy_object_id)
{
	ENEMY_OBJECT_Type* enemy = &enemy_list[0];
	int2 i;

	for (i = 0; i < enemy_count; i++)
	{
		if (enemy->enemy_type == GAME_PIECE_Hatchling)
		{
			if ((enemy->object->flags & FLAG_VA) == FLAG_VA)
			{
				enemy->object->flags = FLAG_NONE;
			}
			else if (enemy->object->id == enemy_object_id)
			{
				enemy->object->flags |= (FLAG_VISIBLE | FLAG_ACTIVE);
				enemy->hold = false;
				enemy->delay = 15;
			}
		}

		enemy++;
	}
}
/*
void ReleaseAllHolds(void)
{
	ENEMY_OBJECT_Type* enemy = &enemy_list[0];
	int2 i;

	for (i = 0; i < enemy_count; i++)
	{
		if (enemy->enemy_type == GAME_PIECE_Robot1)
		{
			if (enemy->hold == true)
			{
				enemy->object->flags |= (FLAG_VISIBLE | FLAG_ACTIVE);
				enemy->hold = false;
				enemy->delay = 20;
				enemy->move_history_index = 0;;

				break;
			}
		}

		enemy++;
	}
}
*/
void Teleport(int2 enemy_object_id) // could make use of the FindEnemyObject
{
	ENEMY_OBJECT_Type* enemy = &enemy_list[0];
	int2 i;

	for (i = 0; i < enemy_count; i++)
	{
		if (enemy->object->id == enemy_object_id)
		{
			jumpman.x = enemy->x;
			jumpman.y = enemy->y;
		}

		enemy++;
	}
}

void ResetEnemiesOnDeath(void)
{
	ENEMY_OBJECT_Type* enemy = &enemy_list[0];
	int2 i;

	for (i = 0; i < enemy_count; i++)
	{
		if (enemy == 0)
			continue;
			
		if (enemy->reset_on_death)
		{
			ResetEnemy(enemy);
		}

		enemy++;
	}
}

ENEMY_OBJECT_Type* FindEnemyObject(int2 enemy_object_id)
{
	ENEMY_OBJECT_Type* enemy = &enemy_list[0];
	int2 i;

	for (i = 0; i < enemy_count; i++)
	{
		if (enemy == 0)
			continue;
			
		if (enemy->object->id == enemy_object_id)
		{
			return enemy;
		}

		enemy++;
	}	
	
	return 0;
}


uint1 GetAIType3Move(ENEMY_OBJECT_Type* enemy)
{
	int2   n = 0;
	uint1  temp_direction = DIR_None;
	uint1  preferred_x_move = DIR_None;
	uint1  preferred_y_move = DIR_None;
	int2   x_diff = enemy->x - jumpman.x;
	int2   y_diff = enemy->y - jumpman.y;
	int2   smart_move_chance = RandomNumber(100);
	int2   fifty_fifty_move_chance = RandomNumber(2);

	VerticalAdjustEnemy(enemy);
	CheckMoveEnemy(enemy);

	if (x_diff < -2)
		preferred_x_move = DIR_Right;
	else if (x_diff > 2)
		preferred_x_move = DIR_Left;
	else
		preferred_x_move = DIR_Left + fifty_fifty_move_chance;

	if (y_diff < -2)
		preferred_y_move = DIR_Down;
	else if (y_diff > 2)
		preferred_y_move = DIR_Up;
	else
		preferred_y_move = DIR_Up + fifty_fifty_move_chance;
/*
	if (continue_move_chance)
	{
		if ((direction == DIR_Up) && (move_up_allowed))
			temp_direction = DIR_Up;
		else if ((direction == DIR_Down) && (move_down_allowed))
			temp_direction = DIR_Down;
		else if ((direction == DIR_Left) && (move_left_allowed))
			temp_direction = DIR_Left;
		else if ((direction == DIR_Right) && (move_right_allowed))
			temp_direction = DIR_Right;
	}

	if (abs(x_diff) < abs(y_diff))
	{
			if ((preferred_x_move == DIR_Right) && (move_right_allowed))
				temp_direction = DIR_Right;
			else if ((preferred_x_move == DIR_Left) && (move_left_allowed))
				temp_direction = DIR_Left;
			else if ((preferred_y_move == DIR_Up) && (move_up_allowed))
				temp_direction = DIR_Up;
			else if ((preferred_y_move == DIR_Down) && (move_down_allowed))
				temp_direction = DIR_Down;	
	}
	else
	{
			if ((preferred_y_move == DIR_Up) && (move_up_allowed))
				temp_direction = DIR_Up;
			else if ((preferred_y_move == DIR_Down) && (move_down_allowed))
				temp_direction = DIR_Down;	
			else if ((preferred_x_move == DIR_Right) && (move_right_allowed))
				temp_direction = DIR_Right;
			else if ((preferred_x_move == DIR_Left) && (move_left_allowed))
				temp_direction = DIR_Left;
	}

	else if (preferred_x_move && preferred_y_move)
	{
		if (RandomNumber(2))
		{
			if ((preferred_x_move == DIR_Right) && (move_right_allowed))
				temp_direction = DIR_Right;
			else if ((preferred_x_move == DIR_Left) && (move_left_allowed))
				temp_direction = DIR_Left;
		}
		else
		{
			if ((preferred_y_move == DIR_Up) && (move_up_allowed))
				temp_direction = DIR_Up;
			else if ((preferred_y_move == DIR_Down) && (move_down_allowed))
				temp_direction = DIR_Down;	
		}
	}
	else if (preferred_x_move )
	{
			if ((preferred_x_move == DIR_Right) && (move_right_allowed))
				temp_direction = DIR_Right;
			else if ((preferred_x_move == DIR_Left) && (move_left_allowed))
				temp_direction = DIR_Left;
	}
	else if (preferred_y_move)
	{
			if ((preferred_y_move == DIR_Up) && (move_up_allowed))
				temp_direction = DIR_Up;
			else if ((preferred_y_move == DIR_Down) && (move_down_allowed))
				temp_direction = DIR_Down;	
	}
*/
/*
	if (direction == DIR_Up)
	{
		if ((preferred_x_move == DIR_Right) && (move_right_allowed))
			temp_direction = DIR_Right;
		else if ((preferred_x_move == DIR_Left) && (move_left_allowed))
			temp_direction = DIR_Left;
		else if (move_up_allowed && continue_move_chance)
			temp_direction = DIR_Up;
		else if (move_down_allowed)
			temp_direction = DIR_Down;
	}
	else if (direction == DIR_Down)
	{
		if ((preferred_x_move == DIR_Right) && (move_right_allowed))
			temp_direction = DIR_Right;
		else if ((preferred_x_move == DIR_Left) && (move_left_allowed))
			temp_direction = DIR_Left;
		else if (move_down_allowed && continue_move_chance)
			temp_direction = DIR_Down;
		else if (move_up_allowed)
			temp_direction = DIR_Up;
	}
	else if (direction == DIR_Left)
	{
		if ((preferred_y_move == DIR_Up) && (move_up_allowed))
			temp_direction = DIR_Up;
		else if ((preferred_y_move == DIR_Down) && (move_down_allowed))
			temp_direction = DIR_Down;
		else if (move_left_allowed && continue_move_chance)
			temp_direction = DIR_Left;
		else if (move_right_allowed)
			temp_direction = DIR_Right;
	}
	else if (direction == DIR_Right)
	{
		if ((preferred_y_move == DIR_Up) && (move_up_allowed))
			temp_direction = DIR_Up;
		else if ((preferred_y_move == DIR_Down) && (move_down_allowed))
			temp_direction = DIR_Down;
		else if (move_right_allowed && continue_move_chance)
			temp_direction = DIR_Right;
		else if (move_left_allowed)
			temp_direction = DIR_Left;
	}
*/
	if (direction == DIR_Up)
	{
		if ((((move_right_allowed) || (move_left_allowed)) && (enemy->direction_count > 4)) || (!move_up_allowed))
		{
			// chose new direction
			if (smart_move_chance)
			{		
				if ((preferred_x_move == DIR_Right) && (move_right_allowed))
					temp_direction = DIR_Right;
				else if ((preferred_x_move == DIR_Left) && (move_left_allowed))
					temp_direction = DIR_Left;
				else if (move_up_allowed )//&& fifty_fifty_move_chance)
					temp_direction = DIR_Up;
				else if (move_down_allowed)
					temp_direction = DIR_Down;
			}
		}
		else if (move_up_allowed)
		{
			temp_direction = DIR_Up;
		}
	}
	else if (direction == DIR_Down)
	{
		if ((((move_right_allowed) || (move_left_allowed)) && (enemy->direction_count > 4)) || (!move_down_allowed))
		{
			// chose new direction
			if (smart_move_chance)
			{		
				if ((preferred_x_move == DIR_Right) && (move_right_allowed))
					temp_direction = DIR_Right;
				else if ((preferred_x_move == DIR_Left) && (move_left_allowed))
					temp_direction = DIR_Left;
				else if (move_down_allowed )//&& fifty_fifty_move_chance)
					temp_direction = DIR_Down;
				else if (move_up_allowed)
					temp_direction = DIR_Up;
			}
		}
		else if (move_down_allowed)
		{
			temp_direction = DIR_Down;
		}
	}
	else if (direction == DIR_Left)
	{
		if ((((move_up_allowed) || (move_down_allowed)) && (enemy->direction_count > 4)) || (!move_left_allowed))
		{
			// chose new direction
			if (smart_move_chance)
			{		
				if ((preferred_y_move == DIR_Up) && (move_up_allowed))
					temp_direction = DIR_Up;
				else if ((preferred_y_move == DIR_Down) && (move_down_allowed))
					temp_direction = DIR_Down;
				else if (move_left_allowed )//&& fifty_fifty_move_chance)
					temp_direction = DIR_Left;
				else if (move_right_allowed)
					temp_direction = DIR_Right;
			}
		}
		else if (move_left_allowed)
		{
			temp_direction = DIR_Left;
		}
	}
	else if (direction == DIR_Right)
	{
		if ((((move_up_allowed) || (move_down_allowed)) && (enemy->direction_count > 4)) || (!move_right_allowed))
		{
			// chose new direction
			if (smart_move_chance)
			{		
				if ((preferred_y_move == DIR_Up) && (move_up_allowed))
					temp_direction = DIR_Up;
				else if ((preferred_y_move == DIR_Down) && (move_down_allowed))
					temp_direction = DIR_Down;
				else if (move_right_allowed )//&& fifty_fifty_move_chance)
					temp_direction = DIR_Right;
				else if (move_left_allowed)
					temp_direction = DIR_Left;
			}
		}
		else if (move_right_allowed)
		{
			temp_direction = DIR_Right;
		}
	}
	else if (direction == DIR_None)
	{
		if (move_allowed[direction] != direction)
		{
			
		// Let's loop a max of 32 times to find a new direction
		while (n++ < 32)
		{
			temp_direction = move_allowed[RandomNumber(4) + 1];
			if (temp_direction != DIR_None)
				break;
		}
		}
	}

			//	if ((direction != temp_direction) && (temp_direction != DIR_None))
			//		enemy->direction_count = 0;

#if 0
	if (temp_direction == DIR_None)
	{
		if (move_allowed[direction] != direction)
		{
			
		// Let's loop a max of 32 times to find a new direction
		while (n++ < 32)
		{
			temp_direction = move_allowed[RandomNumber(4) + 1];
			if (temp_direction != DIR_None)
				break;
		}
		}
	}
#endif



	return temp_direction;
}