// Header File
// Created 11/21/2010; 7:17:06 AM

//******************************************************************************
//
// Project:         Jumpman 89
//
// file name:       jm_levels.h
//
// author:          David Randall "Ranman" 
//                  derandall99@gmail.com
//
// notes::          Jumpman was originally created by Randy Glover for the
//                  Atari 400, Commodore 64, IBM PCjr, and Apple II in 1982
//
//******************************************************************************


#ifndef _jm_levels_h
#define _jm_levels_h

#include "BasicTypes.h"

#include "jm_objects.h"

#define FLAG_NONE                 0x00

// standard object flags
#define FLAG_SCRIPT               0x01
#define FLAG_EXTENDED             0x01
#define FLAG_WAYPOINT             0x02
#define FLAG_ACTIVE               0x04
#define FLAG_VISIBLE              0x08
#define FLAG_MOVING               0x10
#define FLAG_HOLD                 0x20
#define FLAG_FALLING              0x40
#define FLAG_PICKUP               0x80

// script only flags
#define FLAG_COMMAND              0x01
#define FLAG_TOGGLE               0x02 // NOTE: can also be used for openning/closeing doors
#define FLAG_LETTER               0x04
#define FLAG_BONUS                0x08
#define FLAG_OPEN                 0x10
#define FLAG_CLOSE                0x20
#define FLAG_reserved1            0x40
#define FLAG_reserved2            0x80

// direction flags
#define FLAG_DIR_UP               0x01
#define FLAG_DIR_DOWN             0x02
#define FLAG_DIR_LEFT             0x04
#define FLAG_DIR_RIGHT            0x08
#define FLAG_DIR_UPLEFT           0x10
#define FLAG_DIR_UPRIGHT          0x20
#define FLAG_DIR_DOWNLEFT         0x40
#define FLAG_DIR_DOWNRIGHT        0x80
#define FLAG_DIR_UPDOWN           (FLAG_DIR_UP     | FLAG_DIR_DOWN)
#define FLAG_DIR_LEFTRIGHT        (FLAG_DIR_LEFT   | FLAG_DIR_RIGHT)
#define FLAG_DIR_UPDOWNLEFTRIGHT  (FLAG_DIR_UPDOWN | FLAG_DIR_LEFTRIGHT)
#define FLAG_DIR_DIAG             (FLAG_DIR_UPLEFT | FLAG_DIR_UPRIGHT | FLAG_DIR_DOWNLEFT | FLAG_DIR_DOWNRIGHT)

#define FLAG_D_U     FLAG_DIR_UP
#define FLAG_D_D     FLAG_DIR_DOWN
#define FLAG_D_L     FLAG_DIR_LEFT
#define FLAG_D_R     FLAG_DIR_RIGHT
#define FLAG_D_UL    FLAG_DIR_UPLEFT
#define FLAG_D_UR    FLAG_DIR_UPRIGHT
#define FLAG_D_DL    FLAG_DIR_DOWNLEFT
#define FLAG_D_DR    FLAG_DIR_DOWNRIGHT
#define FLAG_D_UD    FLAG_DIR_UPDOWN
#define FLAG_D_LR    FLAG_DIR_UPDOWN
#define FLAG_D_UDLR  FLAG_DIR_UPDOWNLEFTRIGHT
#define FLAG_D_DIAG  FLAG_DIR_DIAG
#define FLAG_D_ALL   FLAG_DIR_DIAG

#define FLAG_V     (FLAG_VISIBLE)
#define FLAG_A     (FLAG_ACTIVE)
#define FLAG_S     (FLAG_SCRIPT)
#define FLAG_X     (FLAG_EXTENDED)
#define FLAG_W     (FLAG_WAYPOINT)
#define FLAG_VA    (FLAG_VISIBLE | FLAG_ACTIVE)
#define FLAG_AS    (FLAG_ACTIVE | FLAG_SCRIPT)
#define FLAG_VP    (FLAG_VISIBLE | FLAG_ACTIVE)
#define FLAG_PH    (FLAG_PICKUP  | FLAG_PICKUP)
#define FLAG_VAF   (FLAG_VISIBLE | FLAG_ACTIVE | FLAG_FALLING)
#define FLAG_VAS   (FLAG_VISIBLE | FLAG_ACTIVE | FLAG_SCRIPT)
#define FLAG_VAW   (FLAG_VISIBLE | FLAG_ACTIVE | FLAG_WAYPOINT)
#define FLAG_VAM   (FLAG_VISIBLE | FLAG_ACTIVE | FLAG_MOVING)
#define FLAG_VAF   (FLAG_VISIBLE | FLAG_ACTIVE | FLAG_FALLING)
#define FLAG_VAH   (FLAG_VISIBLE | FLAG_ACTIVE | FLAG_HOLD)
#define FLAG_VAP   (FLAG_VISIBLE | FLAG_ACTIVE | FLAG_PICKUP)
#define FLAG_VAFH  (FLAG_VISIBLE | FLAG_ACTIVE | FLAG_FALLING | FLAG_HOLD)
#define FLAG_VAHW  (FLAG_VISIBLE | FLAG_ACTIVE | FLAG_HOLD | FLAG_WAYPOINT)
#define FLAG_VAMW  (FLAG_VISIBLE | FLAG_ACTIVE | FLAG_MOVING | FLAG_WAYPOINT)
#define FLAG_VAMWH (FLAG_VISIBLE | FLAG_ACTIVE | FLAG_MOVING | FLAG_WAYPOINT | FLAG_HOLD)
#define FLAG_VAMX  (FLAG_VISIBLE | FLAG_ACTIVE | FLAG_MOVING | FLAG_EXTENDED)
#define FLAG_VAMXH (FLAG_VISIBLE | FLAG_ACTIVE | FLAG_MOVING | FLAG_EXTENDED | FLAG_HOLD)

#define FLAG_MOVABLE (FLAG_MOVING | FLAG_FALLING)

#define TOGGLE_BIT_Girder          0x01
#define TOGGLE_BIT_Ladder          0x02
#define TOGGLE_BIT_Rope_Up         0x04
#define TOGGLE_BIT_Rope_Down       0x08
#define TOGGLE_BIT_Bomb            0x10
#define TOGGLE_BIT_reserved1       0x20
#define TOGGLE_BIT_reserved2       0x40
#define TOGGLE_BIT_Enabled         0x80

typedef enum
{
	SPECIAL_ABILITY_None         = 0,
	SPECIAL_ABILITY_Gunfire      = 1,
	SPECIAL_ABILITY_Spear        = 2,
	SPECIAL_ABILITY_Illuminate   = 3,
	SPECIAL_ABILITY_Hotfoot      = 4,
	SPECIAL_ABILITY_Total,
} SPECIAL_ABILITY_Type;

typedef enum
{
	BULLET_ACTION_None           = 0,
	BULLET_ACTION_Vertical       = 1,
	BULLET_ACTION_HorzAndVert    = 2,
	BULLET_ACTION_Horizontal     = 3,
	BULLET_ACTION_Bounce_Left    = 4,
	BULLET_ACTION_Bounce_Up      = 5,
	BULLET_ACTION_Bounce_Right   = 6,
	BULLET_ACTION_Throw          = 7,
	BULLET_ACTION_Total
} BULLET_ACTION_Type;


#define OBJECTS_PER_LEVEL_MAX 165
#define EXTERNAL_LEVEL_BUFFER_SIZE (30*1024UL + 256)


typedef struct
{
	char name[24]; // including the NULL
	OBJECT_Type object_list[OBJECTS_PER_LEVEL_MAX];
	int2 bonus_points;
	int1 collectable_required;
	int1 wind_speed;
	int1 toggle_bits; // 
	int1 activate_required;
	int1 activate_script_id;
	int1 level_number;
	bool wrap_mode;
	bool final_level;
} LEVEL_Type;



/*

Level design rules & restrictions
  1:  Maximum of 165 unique objects can be defined per level
  2:  Object 'x' value must be between 0 and 39 (lcd.x = object.x * 4)
  3:  Object 'y' value must be between 0 and 88 (lcd.y = object.y)
  3:  Object 'length' value must be between 0 and 88 (lcd.length = object.length * 4)
  4:  Jumpman must be object index 0
  5:  24 enemy objects allowed
  6:  16 bullet objects allowed
  7:  16 moving platform obects allowed
  8:  Moving platform length is limited to 7 (3 least significant bits)
  9:  Moving platforms with waypoints: the waypoint root ID must be <= 31 (5 most significant bits)
 10:  Waypoint root object must point to the first waypoint list obect
 11:  Waypoint list may contain up to 32 waypoints (2 waypoints per waypoint list object)
 11:  Waypoint list objects must be contiguous for each waypoint root

*/

extern bool level_complete;
extern int2 level_number;
extern LEVEL_Type current_level;
extern LEVEL_Type level[];
extern uint1* external_level_buffer;

extern uint1 level_toggle_bits;
extern uint4 level_toggle_girders_count;
extern uint4 level_toggle_ladders_count;
extern uint4 level_toggle_rope_ups_count;
extern uint4 level_toggle_rope_down_count;
extern uint4 level_toggle_bombs_count;


void DisplayNextLevel(void);
void InitLevel(void);
void LoadLevelPack();
void DrawSelectScreen(void);

#endif
